////////////////////////////////////////////////////////////////////////////////
// Main unit of the GF2Demo project
// Unit of the HIT package for DELPHI 4/5
//------------------------------------------------------------------------------
// Copyright (c) 2000, Marcel Martin. All rights reserved.
////////////////////////////////////////////////////////////////////////////////
unit GF2DemoMain;

interface
{$A+,H+}

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, ExtCtrls, Buttons, Menus;

type
  TForm1 = class(TForm)
    Panel1: TPanel;
    Memo1: TMemo;
    AbortBtn: TSpeedButton;
    ClearBtn: TSpeedButton;
    CloseBtn: TSpeedButton;
    HelpBtn: TSpeedButton;
    MenuBtn: TSpeedButton;
    PopupMenu1: TPopupMenu;
    PopItemA: TMenuItem;
    PopItemB: TMenuItem;
    PopItemC: TMenuItem;
    PopItemD: TMenuItem;
    PopItemE: TMenuItem;
    PopItemF: TMenuItem;
    PopItemG: TMenuItem;
    PopItemH: TMenuItem;
    PopItemI: TMenuItem;
    procedure AbortBtnClick(Sender: TObject);
    procedure ClearBtnClick(Sender: TObject);
    procedure CloseBtnClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure HelpBtnClick(Sender: TObject);
    procedure MenuBtnClick(Sender: TObject);
    procedure PopItemAClick(Sender: TObject);
    procedure PopItemBClick(Sender: TObject);
    procedure PopItemCClick(Sender: TObject);
    procedure PopItemDClick(Sender: TObject);
    procedure PopItemEClick(Sender: TObject);
    procedure PopItemFClick(Sender: TObject);
    procedure PopItemGClick(Sender: TObject);
    procedure PopItemHClick(Sender: TObject);
    procedure PopItemIClick(Sender: TObject);
    procedure Run(Sender: TObject; Num: Char);
    procedure RunA(Sender: TObject);
    procedure RunB(Sender: TObject);
    procedure RunC(Sender: TObject);
    procedure RunD(Sender: TObject);
    procedure RunE(Sender: TObject);
    procedure RunF(Sender: TObject);
    procedure RunG(Sender: TObject);
    procedure RunH(Sender: TObject);
    procedure RunI(Sender: TObject);
  end;

var Form1: TForm1;


implementation
//==============================================================================
{$C-,I-,O+,Q-,R-,W-}
{$R *.DFM}
//==============================================================================
uses
  Integers,
  GF2,
  GF2N_EllipticCurves,
  GF2N_ECSubfields,
  GF2N_GNBFieldPolys;

const
  CHECKMSG : array[Boolean] of string[5] = ('ERROR','ok');


////////////////////////////////////////////////////////////////////////////////
// TForm1 class
////////////////////////////////////////////////////////////////////////////////

//==============================================================================
// Abort the running of some HIT routines
//==============================================================================
procedure TForm1.AbortBtnClick(Sender: TObject);
begin
  HITAbort;
end;

//==============================================================================
// Clear the editor
//==============================================================================
procedure TForm1.ClearBtnClick(Sender: TObject);
begin
  Memo1.Lines.Clear;
end;

//==============================================================================
// Terminate the application
//==============================================================================
procedure TForm1.CloseBtnClick(Sender: TObject);
begin
  Close;
end;

//==============================================================================
// Create the form
//==============================================================================
procedure TForm1.FormCreate(Sender: TObject);
begin
  Application.HintColor := clYELLOW;

  //----- String format
  //HITStrFormat.SpaceAsPlus := false; //'+' instead of a space
  //HITStrFormat.LowerCased := false; //For hexadecimal strings
end;

//==============================================================================
// Open the HIT.HLP file
//==============================================================================
procedure TForm1.HelpBtnClick(Sender: TObject);
begin
  Application.HelpCommand(HELP_FINDER,0);
end;

////////////////////////////////////////////////////////////////////////////////
// Example A
////////////////////////////////////////////////////////////////////////////////

//==============================================================================
// Generate and display a random polynomial over GF(2)
// Polynomials over GF(2) are also called "Binary polynomials". The coefficients
// are always either 0 or 1, and the computations are done modulo 2. That's why
// it is very convenient to represent them as bit sequences.
//==============================================================================
procedure TForm1.RunA(Sender: TObject);
  const INFO = '  (Parity is equivalent to "Weight and 1")';
  var
    P : TGF2Poly;
    A : TInteger;
begin
  P := nil;
  A := nil;
  with Memo1.Lines do
  try
    //----- Create P equal to 0
    P := TGF2Poly.Create;
    //----- Random polynomial over GF(2) of (at most) degree 128
    GF2Rnd(P,128,false);

    //===== Display the polynomial

    //----- Binary representation: One bit = One coefficient
    Add('P (Binary representation)');
    Add('-------------------------');
    Add(GF2BinStr(P,8)); //8-bit blocks
    Add('Degree  = ' + IntToStr(P.Degree));
    Add('Density = ' + FloatToStrF(P.Density,ffFixed,18,3));
    Add('Index   = ' + IntToStr(P.Index));
    Add('Parity  = ' + IntToStr(Ord(P.Parity)) + INFO);
    Add('Weight  = ' + IntToStr(P.Weight));
    Add('');

    //----- Hexa representation: Same as binary representation but the
    //      values are packed in hexadecimal values
    Add('P (Hexadecimal representation)');
    Add('------------------------------');
    Add(GF2HexStr(P,2)); //2-digit blocks
    Add('');

    //----- Display the degrees of coefficients equal to 1
    Add('P (Degrees of coefs equal to 1)');
    Add('-------------------------------');
    Add(GF2DegsStr(P));
    Add('');

    //----- Conventional representation of a polynomial
    Add('P (Common representation)');
    Add('-------------------------');
    Add(GF2PolyStr(P,'X'));
    Add('');

    //----- "Decimal" representation
    //      To display a polynomial to the base 10, you first need to
    //      cast it to TInteger type
    A := TInteger.Create;
    GF2PolyToInt(P,A); //Cast P to TInteger
    Add('P (Decimal representation)');
    Add('--------------------------');
    HITStrFormat.Signed := false;
    Add(IStr(A,10,5)); //5-digit blocks
    HITStrFormat.Signed := true;
    Add('');

  finally
    //----- Free the polynomial
    P.Free;
    //----- Free the integer
    A.Free;
  end;
end;

////////////////////////////////////////////////////////////////////////////////
// Example B
////////////////////////////////////////////////////////////////////////////////

//==============================================================================
// Factorization of polynomials over GF(2)
//==============================================================================
procedure TForm1.RunB(Sender: TObject);
  //----- Constants to display polynomials (4-digit blocks)
  const BLOCKS = 4;
  var
    P, Q : TGF2Poly;
    F    : TGF2PolyVector;
    Test : Boolean;
    i    : Integer;
begin
  P := nil;
  Q := nil;
  F := nil;
  with Memo1.Lines do
  try
    //----- Create 2 polynomials equal to 0
    P := TGF2Poly.Create;
    Q := TGF2Poly.Create;

    //----- Create an empty vector
    F := TGF2PolyVector.Create;

    //----- Generate a composite polynomial of degree = 768
    GF2Rnd(P,256);
    for i := 1 to 2 do
    begin
      GF2Rnd(Q,256); //Random, degree equal to 256
      GF2Mul(P,Q);   //P := P * Q
    end;
    Add('P =  ' + GF2HexStr(P,BLOCKS));
    Add('P.Degree = ' + IntToStr(P.Degree));
    Add('');

    //===== Factorization
    Add('List of the irreducible factors of P');
    Add('------------------------------------');
    GF2Split(P,F,true); //F <- Irreducible factors of P
    F.Sort;
    Q.Set1;
    for i := 0 to F.Size-1 do
    begin
      //----- Build Q for checking
      GF2Mul(Q,F[i]);
      //----- Display facor #i
      Add('F = ' + GF2HexStr(F[i],BLOCKS));
    end;
    Add('');

    //===== Here, Q should be equal to P without squared factors
    GF2SqrFree(P);
    GF2Add(P,Q);
    Add('Check product -> ' + CHECKMSG[P.Is0]);

    //===== Check the irreducibility of the factors
    Test := true;
    for i := 0 to F.Size-1 do
      if not F[i].IsIrreducible then
      begin
        Add('');
        Add('F = ' + GF2HexStr(F[i],BLOCKS) + ' is not irreducible');
        Test := false;
        Break;
      end;
    Add('Check irreducibility -> ' + CHECKMSG[Test]);
    Add('');

  finally
    P.Free;
    Q.Free;
    F.Free;
  end;
end;

////////////////////////////////////////////////////////////////////////////////
// Example C
////////////////////////////////////////////////////////////////////////////////

//==============================================================================
// Matrices over GF(2)
//==============================================================================
procedure TForm1.RunC(Sender: TObject);
  //--------------------------------------------------------------------------
  procedure EditMat(const Title: string; M: TGF2SqrMatrix);
    const C : array[Boolean] of AnsiChar = ('0','1');
    var
      S         : AnsiString;
      i, j, dim : Longint;
  begin
    Memo1.Lines.Add(Title);
    Memo1.Lines.Add(StringOfChar('-',Length(Title)));
    dim := M.Dimension;
    for i := 0 to dim-1 do
    begin
      S := '';
      for j := 0 to dim-1 do S := S + C[M[i,j]] + ' ';
      Memo1.Lines.Add(S);
    end;
    Memo1.Lines.Add('');
  end;
  //--------------------------------------------------------------------------
  var
    M, N   : TGF2SqrMatrix;
    P      : TGF2Poly;
    dim, i : Longint;
begin
  P := nil;
  M := nil;
  N := nil;
  with Memo1.Lines do
  try
    //----- Create a null polynomial
    P := TGF2Poly.Create;

    //----- Create an empty matrix (dimension = 0)
    M := TGF2SqrMatrix.Create;

    //----- Get a 17 x 17 random invertible matrix
    dim := 17;
    M.Dimension := dim;
    repeat
      for i := 0 to dim-1 do
      begin
        GF2Rnd(P,dim-1,false);
        M.SetRow(i,P);
      end;
    until M.Determinant;
    EditMat('Matrix M',M);

    //----- Create matrix N equal to M^(-1)
    N := M.Clone;
    GF2Inv(N);
    EditMat('Matrix M^(-1)',N);

    //----- M := M * N (Should be the identity matrix)
    GF2MulByRight(M,N);
    EditMat('Identity',M);

  finally
    //----- Free the matrices
    M.Free;
    N.Free;
    //----- Free the polynomial
    P.Free;
  end;
end;

////////////////////////////////////////////////////////////////////////////////
// Example D
////////////////////////////////////////////////////////////////////////////////

//==============================================================================
// Operations over GF(2^N)
//==============================================================================
procedure TForm1.RunD(Sender: TObject);
  //----- Constants to display polynomials (4-digit blocks)
  const BLOCKS = 4;
  var
    P, Q, R, S : TGF2Poly;
    gf         : TGF2N;
    E          : TInteger;
    TickCount  : Int64;
    cd         : array[1..4] of Extended;
begin
  E := nil;
  P := nil;
  Q := nil;
  R := nil;
  S := nil;
  gf := nil;
  with Memo1.Lines do
  try
    //----- Create an integer equal to 0
    E := TInteger.Create;

    //----- Create 4 polynomials equal to 0
    P := TGF2Poly.Create;
    Q := TGF2Poly.Create;
    R := TGF2Poly.Create;
    S := TGF2Poly.Create;

    //----- Create and initialize a Galois field
    gf := TGF2N.Create(180,btStandard);
    gf.CheckParam := false;

    //===== Distributivity: P(Q + R) = PQ + PR
    //----- Random P, Q, R
    gf.Rnd(P);
    gf.Rnd(Q);
    gf.Rnd(R);
    TickCount := HITTimer;
    //----- Compute S := P(Q + R)
    GF2Set(S,Q);
    gf.Add(S,R);
    gf.Mul(S,P);
    //----- Compute P := PQ + PR
    gf.Mul(R,P);
    gf.Mul(P,Q);
    gf.Add(P,R);
    cd[1] := HITTimer(TickCount);
    //------ Display
    Add('Distributivity test: P(Q + R) = PQ + PR ?');
    Add('-----------------------------------------');
    Add('(1) = ' + GF2HexStr(S,BLOCKS));
    Add('(2) = ' + GF2HexStr(P,BLOCKS));
    gf.Add(P,S);
    Add('Check -> ' + CHECKMSG[P.Is0]);
    Add('');

    //===== Square root: (P^(1/2))^2 = P
    //----- Random P
    gf.Rnd(P);
    //----- Compute Q := (P^(1/2))^2
    TickCount := HITTimer;
    GF2Set(Q,P);
    gf.SqrRoot(Q);
    gf.Sqr(Q);
    cd[2] := HITTimer(TickCount);
    //------ Display
    Add('Square root test: P = (P^(1/2))^2 ?');
    Add('-----------------------------------');
    Add('(1) = ' + GF2HexStr(P,BLOCKS));
    Add('(2) = ' + GF2HexStr(Q,BLOCKS));
    gf.Add(P,Q);
    Add('Check -> ' + CHECKMSG[P.Is0]);
    Add('');

    //===== Inversion: Q(P/Q) = P
    //----- Random P, Q
    gf.Rnd(P);
    repeat
      gf.Rnd(Q);
    until not Q.Is0;
    //----- Compute Q := Q(P/Q)
    TickCount := HITTimer;
    GF2Set(R,Q);
    gf.Inv(Q);
    gf.Mul(Q,P);
    gf.Mul(Q,R);
    cd[3] := HITTimer(TickCount);
    //------ Display
    Add('Inversion test: P = Q * (P * Q^(-1)) ?');
    Add('--------------------------------------');
    Add('(1) = ' + GF2HexStr(P,BLOCKS));
    Add('(2) = ' + GF2HexStr(Q,BLOCKS));
    gf.Add(P,Q);
    Add('Check -> ' + CHECKMSG[P.Is0]);
    Add('');

    //===== Power: 1/P = P^(2^N - 2)
    //----- Random P
    repeat
      gf.Rnd(P);
    until not P.Is0;
    //----- Compute Q := P^(2^N - 2)
    GF2Set(Q,P);
    gf.Order(E); //Number of elements of GF(2^N) = 2^N
    ISub(E,2);
    TickCount := HITTimer;
    gf.Pow(Q,E);
    cd[4] := HITTimer(TickCount);
    //----- Compute P := 1/P
    gf.Inv(P);
    //------ Display
    Add('Power test: 1/P = P^(2^N - 2) ?');
    Add('-------------------------------');
    Add('(1) = ' + GF2HexStr(P,BLOCKS));
    Add('(2) = ' + GF2HexStr(Q,BLOCKS));
    gf.Add(P,Q);
    Add('Check -> ' + CHECKMSG[P.Is0]);
    Add('');

    //===== Running times
    Add('Running times (microseconds):');
    Add('-----------------------------');
    Add('Distributivity test .. ' + FloatToStrF(cd[1]*1000,ffFixed,18,0));
    Add('Square root test ..... ' + FloatToStrF(cd[2]*1000,ffFixed,18,0));
    Add('Inversion test ....... ' + FloatToStrF(cd[3]*1000,ffFixed,18,0));
    Add('Power test ........... ' + FloatToStrF(cd[4]*1000,ffFixed,18,0));
    Add('');

  finally
    //----- Free the field
    gf.Free;
    //----- Free the polynomials
    P.Free;
    Q.Free;
    R.Free;
    S.Free;
    //----- Free the integer
    E.Free;
  end;
end;

////////////////////////////////////////////////////////////////////////////////
// Example E
////////////////////////////////////////////////////////////////////////////////

//==============================================================================
// Change-of-basis over GF(2^N)
//==============================================================================
procedure TForm1.RunE(Sender: TObject);
  //----- Constants to display polynomials (4-digit blocks)
  const BLOCKS = 4;
  //--------------------------------------------------------------------------
  // U := F(U) over GF(2^N)
  // Here, F(U) = U/(U^6 + 1), but it may be any
  //--------------------------------------------------------------------------
  procedure Func(U: TGF2Poly; gf: TGF2N);
    var V : TGF2Poly;
  begin
    V := U.Clone;
    try
      gf.Pow(U,6);
      gf.Add1(U); //And not "GF2Add1" because of possible normal basis
      gf.Inv(U);
      gf.Mul(U,V);
    finally
      V.Free;
    end;
  end;
  //--------------------------------------------------------------------------
  var
    P, Q, R, U1, U2 : TGF2Poly;
    gf1, gf2        : TGF2N;
    M               : TGF2SqrMatrix;
    Type1, Type2    : TGF2NBasisType;
    Deg             : Longint;
    TickCount       : Int64;
    c               : Extended;
begin
  P := nil;
  Q := nil;
  R := nil;
  U1 := nil;
  U2 := nil;
  gf1 := nil;
  gf2 := nil;
  M := nil;
  with Memo1.Lines do
  try
    //----- Allow to abort calculations
    HITBreakWith(Application.ProcessMessages);
    AbortBtn.Enabled := HITBreakOn;

    //----- Create 5 polynomials equal to 0
    P := TGF2Poly.Create;
    Q := TGF2Poly.Create;
    R := TGF2Poly.Create;
    U1 := TGF2Poly.Create;
    U2 := TGF2Poly.Create;

    //----- Create 2 fields
    gf1 := TGF2N.Create;
    gf2 := TGF2N.Create;

    //----- Create an empty matrix
    M := TGF2SqrMatrix.Create;

    //===== Initialize Galois fields
    //      For a given N, it exists only ONE Galois field GF(2^N). Here,
    //      "2 Galois fields" means "2 instances of TGF2N with same degree but
    //      with different basis types and/or different field polynomials".
    //----- Random degree in 80..100
    Deg := (HITRnd32 mod 21) + 80;
    //----- Random basis types: Standard or Normal
    Type1 := TGF2NBasisType(HITRnd32 and 1);
    Type2 := TGF2NBasisType(HITRnd32 and 1);
    (*
    // With the following parameters, the root computation is anormally fast
    Deg := 127;
    Type1 := btStandard;
    Type2 := btNormal;
    *)
    if Type1 = Type2 then
    begin
      case Type1 of
        btStandard: GF2RndIrreducible(P,Deg);
        btNormal:   GF2RndNormal(P,Deg);
      end;
      gf1.SetField(P,Type1,false);
    end
    else
    begin
      gf1.SetField(Deg,Type1);
      gf1.GetFieldPoly(P);
    end;
    //----- Display field info
    gf2.SetField(Deg,Type2);
    gf2.GetFieldPoly(Q);
    Add('Fields');
    Add('------');
    Add('Field degree         = ' + IntToStr(P.Degree));
    Add('Field polynomial P   = ' + GF2HexStr(P,0));
    Add('GF(2^N)/P basis type = ' + gf1.BasisTypeStr);
    Add('Field polynomial Q   = ' + GF2HexStr(Q,0));
    Add('GF(2^N)/Q basis type = ' + gf2.BasisTypeStr);
    Add('');

    //===== Root of P over GF(2^N)/Q
    //      !!! THE COMPUTATION CAN BE VERY LENGTHY
    Add('Computing a root of P over GF(2^N)/Q ...');
    //----- Search a random root (it exists P.Degree distinct roots)
    TickCount := HITTimer;
    gf2.FindRoot(R,P);
    c := HITTimer(TickCount);

    Delete(Count-1);
    Add('Root of P over GF(2^N)/Q');
    Add('------------------------');
    Add('R = ' + GF2HexStr(R,4));
    Add('Running time: ' + FloatToStrF(c/1000,ffFixed,18,3) + ' s');
    Add('');

    //----- Check: U := P(R) over GF(2^N)/Q (Should be 0)
    GF2Set(U1,P);
    gf2.Compose(U1,R);
    if not U1.Is0 then
    begin
      Add('ERROR -> Invalid root');
      Exit;
    end;

    //----- Initialize a change-of-basis matrix (gf2 <- gf1)
    gf2.MakeImportCoBMatrix(M,R,gf1.BasisType);

    // !!! WARNING
    // Since there exist N distinct roots of P in GF(2^N)/Q, we can build up
    // to N distinct change-of-basis matrices M from GF(2^N)/P to GF(2^N)/Q.
    // If U is a random polynomial of GF(2^N)/P different from 0 and from 1,
    // the value M*U is of course different for each matrix M.
    // What is always verified is F(U) = M^(-1) * F(M*U) for any regular
    // function F with any matrix M among the N possible ones.
    // F(U) indicates computations on U in gf1.
    // F(M*U) indicates computations on the image of U in gf2.

    //===== Random U over gf1 and conversion to gf2
    Add('Random polynomial U');
    Add('-------------------');
    // The random poly must be different from 1 because of the used function:
    // with U1 = 1, the denominator would be 0.
    repeat
      gf1.Rnd(U1);
    until not gf1.Is1(U1);
    Add('U over GF(2^N)/P = ' + GF2HexStr(U1,4));
    Add('');
    Add('Conversion of U from GF(2^N)/P to GF(2^N)/Q');
    Add('-------------------------------------------');
    GF2Set(U2,U1);
    //----- Convert from gf1 to gf2
    GF2MulByLeft(U2,M); //U2 := M * U2
    Add('U over GF(2^N)/Q = ' + GF2HexStr(U2,4));
    Add('');

    //===== Computation over gf1 and gf2
    Add('F(U) = U / (U^6 + 1)');
    Add('--------------------');
    Func(U1,gf1);
    Func(U2,gf2);
    Add('F(U) over GF(2^N)/P = ' + GF2HexStr(U1,4));
    Add('F(U) over GF(2^N)/Q = ' + GF2HexStr(U2,4));
    Add('');

    //===== Change of basis from gf2 to gf1
    //----- Inverse the matrix
    GF2Inv(M);
    //----- Change of basis
    GF2MulByLeft(U2,M);
    Add('Conversion of F(U) from GF(2^N)/Q to GF(2^N)/P');
    Add('----------------------------------------------');
    Add('F(U) over GF(2^N)/P = ' + GF2HexStr(U2,4));
    Add('');
    Add('Check -> ' + CHECKMSG[GF2Equ(U1,U2)]);
    Add('');
  finally
    AbortBtn.Enabled := false;
    HITNoBreak;
    //----- Free all
    P.Free;
    Q.Free;
    R.Free;
    U1.Free;
    U2.Free;
    gf1.Free;
    gf2.Free;
    M.Free;
  end;
end;

////////////////////////////////////////////////////////////////////////////////
// Example F
////////////////////////////////////////////////////////////////////////////////

//==============================================================================
// Elliptic Curves over GF(2^N)
// How to compute a point Q from a point P such that 2P = 2Q
//==============================================================================
procedure TForm1.RunF(Sender: TObject);
  //----- Constant to display polynomials (4-digit blocks)
  const BLOCKS = 4;
  //--------------------------------------------------------------------------
  // Display the point P
  //--------------------------------------------------------------------------
  procedure DisplayPoint(const Title: string; P: TGF2NEPoint; c: TGF2NECurve);
  begin
    with Memo1.Lines do
    begin
      Add(Title + ':');
      Add(StringOfChar('-',Length(Title) + 1));
      if P.IsO then Add('IDENTITY')
      else
      begin
        Add('X = ' + GF2HexStr(P.X,BLOCKS));
        Add('Y = ' + GF2HexStr(P.Y,BLOCKS));
        if not c.OnCurve(P) then
          Add('An error occurred: The point is not on the curve!?');
      end;
      Add('');
    end;
  end;
  //--------------------------------------------------------------------------
  var
    A, B                 : TGF2Poly;
    ec                   : TGF2NECurve;
    P, Q, D1, D2, H1, H2 : TGF2NEPoint;
    TickCount            : Int64;
    cd                   : array[1..6] of Extended;
begin
  A := nil;
  B := nil;
  ec := nil;
  P := nil;
  Q := nil;
  D1 := nil;
  D2 := nil;
  H1 := nil;
  H2 := nil;
  with Memo1.Lines do
  try
    //----- Create 2 polynomials equal to 0
    A := TGF2Poly.Create;
    B := TGF2Poly.Create;

    //----- Create an elliptic curve over GF(2^N)
    ec := TGF2NECurve.Create;

    //----- Initialize the curve (field)
//  ec.SetField(179,btStandard);
    ec.SetField(179,btNormal);
    ec.SwapFieldBasisType;

    // The previous combination "Normal Basis" + "Standard Mode" gives
    // the fastest fields (on average) for elliptic curves

    //----- Initialize the curve (coefs)
    ec.SetRndCoefs;

    //------ Display the curve coefficients
    Add('Curve coefficients and field polynomial:');
    Add('----------------------------------------');
    ec.GetCoefs(A,B);
    Add('A = ' + GF2HexStr(A,BLOCKS));
    Add('B = ' + GF2HexStr(B,BLOCKS));
    ec.GetFieldPoly(A);
    Add('P = ' + GF2HexStr(A,BLOCKS));
    Add('');

    //----- Create 6 points equal to O (identity element)
    P := TGF2NEPoint.Create;
    Q := TGF2NEPoint.Create;
    D1 := TGF2NEPoint.Create;
    D2 := TGF2NEPoint.Create;
    H1 := TGF2NEPoint.Create;
    H2 := TGF2NEPoint.Create;

    // What follows show how to compute a point Q from a point P
    // such that (Q <> P) and (2Q = 2P)

    //----- Random P
    TickCount := HITTimer;
    ec.GetRndPoint(P);
    cd[1] := HITTimer(TickCount);

    //----- Get the only point of order 2 of the curve
    TickCount := HITTimer;
    ec.GetOrder2Point(Q);
    cd[2] := HITTimer(TickCount);

    //----- Compute Q := Q + P (Since 2Q = O, we'll have 2(Q + P) = 2P
    TickCount := HITTimer;
    ec.Add(Q,P);
    cd[3] := HITTimer(TickCount);

    //------ Display P and Q
    DisplayPoint('Random point P',P,ec);
    DisplayPoint('Q = P + (0,SqrRoot(B))',Q,ec);

    //----- Double the 2 points
    D1.Copy(P);
    TickCount := HITTimer;
    ec.Double(D1);
    cd[4] := HITTimer(TickCount);
    D2.Copy(Q);
    TickCount := HITTimer;
    ec.Double(D2);
    cd[5] := HITTimer(TickCount);

    //------ Display (we should have P = Q)
    DisplayPoint('2P',D1,ec);
    DisplayPoint('2Q',D2,ec);

    //----- Compute the 2 halves of 2P (the result should be P and Q)
    TickCount := HITTimer;
    ec.Halves(H1,H2,D1);
    cd[6] := HITTimer(TickCount);

    //------ Display
    DisplayPoint('H1',H1,ec);
    DisplayPoint('H2',H2,ec);

    //===== Check
    Add('Check (2P = 2Q) -> ' + CHECKMSG[D1.Equ(D2)]);
    Add('Check halves -> ' + CHECKMSG[(H1.Equ(P) and H2.Equ(Q)) or
                                      (H2.Equ(P) and H1.Equ(Q))]);
    Add('');

    //===== Running times
    Add('Running times (milliseconds):');
    Add('-----------------------------');
    Add('Random point  ' + FloatToStrF(cd[1],ffFixed,18,3));
    Add('Order-2 point ' + FloatToStrF(cd[2],ffFixed,18,3));
    Add('Addition .... ' + FloatToStrF(cd[3],ffFixed,18,3));
    Add('Double ...... ' + FloatToStrF(cd[4],ffFixed,18,3));
    Add('Double ...... ' + FloatToStrF(cd[5],ffFixed,18,3));
    Add('Halves ...... ' + FloatToStrF(cd[6],ffFixed,18,3));
    Add('');

  finally
    //----- Free the curve
    ec.Free;
    //----- Free the polynomials
    A.Free;
    B.Free;
    //----- Free the points
    P.Free;
    Q.Free;
    D1.Free;
    D2.Free;
    H1.Free;
    H2.Free;
  end;
end;

////////////////////////////////////////////////////////////////////////////////
// Example G
////////////////////////////////////////////////////////////////////////////////

//==============================================================================
// Random Elliptic Curves over GF(2^N) via Subfields
// See the "GF2N_ECSubfields.pas" unit for more information
//==============================================================================
procedure TForm1.RunG(Sender: TObject);
  const
    //----- Constant to display polynomials (4-digit blocks)
    BLOCKS = 4;
    //----- EC multiplications are faster when TRUE
    JACOBIAN_COORDINATES = TRUE;
  //--------------------------------------------------------------------------
  // Display the point P
  //--------------------------------------------------------------------------
  procedure DisplayPoint(const Title: string; P: TGF2NEPoint; c: TGF2NECurve);
  begin
    with Memo1.Lines do
    begin
      Add(Title + ':');
      Add(StringOfChar('-',Length(Title) + 1));
      if P.IsO then Add('IDENTITY')
      else
      begin
        Add('X = ' + GF2HexStr(P.X,BLOCKS));
        Add('Y = ' + GF2HexStr(P.Y,BLOCKS));
        if not c.OnCurve(P) then
          Add('An error occurred: The point is not on the curve!?');
      end;
      Add('');
    end;
  end;
  //--------------------------------------------------------------------------
  var
    gfm, gfd     : TGF2N;
    btm, btd     : TGF2NBasisType;
    A, B, SA, SB : TGF2Poly;
    ec           : TGF2NECurve;
    P, Q         : TGF2NEPoint;
    O, F         : TInteger;
    m, d, e      : Longint;
    Found        : Boolean;
begin
  gfm := nil;
  gfd := nil;
  ec := nil;
  A := nil;
  B := nil;
  SA := nil;
  SB := nil;
  P := nil;
  Q := nil;
  O := nil;
  F := nil;
  with Memo1.Lines do
  try
    HITBreakWith(Application.ProcessMessages);
    //----- Activate the Abort button (if need be)
    AbortBtn.Enabled := HITBreakOn;

    //----- Create 2 fields and 2 polynomials
    gfm := TGF2N.Create;
    gfd := TGF2N.Create;

    //----- Create 4 polynomials
    A := TGF2Poly.Create;
    B := TGF2Poly.Create;
    SA := TGF2Poly.Create;
    SB := TGF2Poly.Create;

    //----- Create 2 integers
    O := TInteger.Create;
    F := TInteger.Create;

    //----- Random field degrees such that (4 <= d <= 12) and (8 <= e <= 64)
    //      d*e < 256 so that we can call SetField without computing field
    //      polynomial
    d := (HITRnd32 mod 9) + 4;
    repeat
      e := (HITRnd32 mod 65) + 8;
      m := d * e;
    until (m < 256) and (IGCD32(d,e) = 1);

    //----- Random basis types: Standard or Normal
    btm := TGF2NBasisType(HITRnd32 and 1);
    btd := TGF2NBasisType(HITRnd32 and 1);

    //----- Set the fields
    gfm.SetField(m,btm);
    gfm.CheckParam := false;
    gfd.SetField(d,btd);
    gfd.CheckParam := false;

    Add('Random field degrees:');
    Add('---------------------');
    Add('d = ' + IntToStr(d));
    Add('m = ' + IntToStr(m) + ' = d * ' + IntToStr(e));
    Add('');
    Add('Random basis types:');
    Add('-------------------');
    Add('GF(2^d) basis type = ' + gfd.BasistypeStr);
    Add('GF(2^m) basis type = ' + gfm.BasistypeStr);
    Add('');

    //----- Try to build an elliptic curve via subfield
    Add('Computing a random curve ...');
    Found := GetSubfieldCurve(A,B,SA,SB,O,F,100000,gfm,gfd);
    Delete(Count-1);

    if not Found then Add('NO CURVE FOUND')
    else
    begin
      Add('Curve order =' + IStr(O,10,0));
      Add('Curve order =' + IStr(F,10,0) + ' * Z (with Z prime)');
      Add('');

      //----- Create an elliptic curve over GF(2^N)
      ec := TGF2NECurve.Create;

      ec.SetField(gfm);
      ec.SetCoefs(A,B);

      Add('Curve coefficients:');
      Add('-------------------');
      Add('A = ' + GF2HexStr(A,BLOCKS));
      Add('B = ' + GF2HexStr(B,BLOCKS));
      Add('');
      Add('Subfield curve coefficients:');
      Add('----------------------------');
      Add('A = ' + GF2HexStr(SA,BLOCKS));
      Add('B = ' + GF2HexStr(SB,BLOCKS));
      Add('');

      //----- Create 2 elliptic curve points
      P := TGF2NEPoint.Create;
      Q := TGF2NEPoint.Create;

      ec.GetRndPoint(P);
      Q.Copy(P);
      DisplayPoint('Random point P',P,ec);

      //----- Random invertible multiplicator
      e := O.BitSize shr 1;
      repeat
        IRnd(F,e,false);
        F.Odd := true
      until IGCD1(F,O);
      Add('Random integer K invertible modulo curve order:');
      Add('-----------------------------------------------');
      Add('K =' + IStr(F,10,3));
      Add('');

      //----- P := P*K
      ec.Mul(P,F,JACOBIAN_COORDINATES);
      DisplayPoint('Q = P*K',P,ec);

      //----- P := P*(K^(-1) modulo curve order)
      IInvMod(F,O);
      ec.Mul(P,F,JACOBIAN_COORDINATES);
      DisplayPoint('R = Q*K^(-1)',P,ec);

      //===== Check
      Add('Check (R = P) -> ' + CHECKMSG[P.Equ(Q)]);
    end;
    Add('');
  finally
    //----- Restore the default status
    HITNoBreak;
    AbortBtn.Enabled := false;
    //----- Free the fields
    gfm.Free;
    gfd.Free;
    //----- Free the integers
    O.Free;
    F.Free;
    //----- Free the polynomials
    A.Free;
    B.Free;
    SA.Free;
    SB.Free;
    //----- Free the curve
    ec.Free;
    //----- Free the points
    P.Free;
    Q.Free;
  end;
end;

////////////////////////////////////////////////////////////////////////////////
// Example H
////////////////////////////////////////////////////////////////////////////////

//==============================================================================
// Factorization of a polynomial over GF(2)
//==============================================================================
procedure TForm1.RunH(Sender: TObject);
  var
    P         : TGF2Poly;
    F         : TGF2PolyVector;
    S         : string;
    i         : Longint;
    TickCount : Int64;
    Chrono    : Extended;
begin
  P := nil;
  F := nil;
  with Memo1.Lines do
  try
    //----- Create a null polynomial
    P := TGF2Poly.Create;

    //----- Create an empty vector
    F := TGF2PolyVector.Create;

    //----- Set the polynomial equal to X^32768 + X (A big polynomial!)
    P[32768] := true;
    P[1] := true;

    //----- Allow to abort calculations
    HITBreakWith(Application.ProcessMessages);
    try
      //----- Activate the Abort button (if need be)
      AbortBtn.Enabled := HITBreakOn;

      Add('Factorization in progress ...');

      //----- Factorization of P
      TickCount := HITTimer;
      GF2Split(P,F,true);
      F.Sort;
      Chrono := HITTimer(TickCount);

    finally
      //----- Restore the default status
      HITNoBreak;
      AbortBtn.Enabled := false;
    end;

    //----- Display the 2192 factors (and free them)
    Delete(Count-1);
    S := '';
    for i := 0 to F.Size-1 do S := S + GF2HexStr(F[i],4) + ', ';
    Add(S);
    Add('');

    //----- Display infos
    Add(IntToStr(F.Size) + ' factors');
    Add('Running time = ' + FloatToStrF(Chrono/1000,ffFixed,18,3) + ' seconds');
    Add('');
  finally
    //----- Free the factors
    F.Free;
    //----- Free the polynomial
    P.Free;
  end;
end;

////////////////////////////////////////////////////////////////////////////////
// Example I
////////////////////////////////////////////////////////////////////////////////

//==============================================================================
// Computation of a Gaussian normal field polynomial
// See the "GF2N_GNBFieldPolys.pas" unit for more information
//==============================================================================
procedure TForm1.RunI(Sender: TObject);
  var
    P            : TGF2Poly;
    N, GaussType : Longint;
    TickCount    : Int64;
    Chrono       : Extended;
begin
  P := nil;
  with Memo1.Lines do
  try
    //----- Create a null polynomial
    P := TGF2Poly.Create;

    //----- Get a random N in 2..255
    repeat
      N := HITRnd32 and $ff;
      GaussType := GaussianNormalBasisType(N);
    until GaussType <> 0;
    Add('Degree = ' + IntToStr(N));
    Add('Type   = ' + IntToStr(GaussType));
    Add('');

    //----- Allow to abort calculations
    HITBreakWith(Application.ProcessMessages);
    try
      //----- Activate the Abort button (if need be)
      AbortBtn.Enabled := HITBreakOn;

      // Unlike the computation method using polynomials over C, the running
      // times of the method used by HIT depends only on the N parameter (and
      // not on both N and GaussType parameters).
      // Here are some running times obtained with an AMD K6-2/450 processor.
      //   N = 63  ->  0.25s
      //   N = 127 ->  3s
      //   N = 159 ->  8s
      //   N = 190 -> 15s
      //   N = 223 -> 28s
      //   N = 255 -> 47s

      Add('Computation in progress ...');

      //----- Compute the polynomial (For types I and II, that's instant)
      TickCount := HITTimer;
      GaussianFieldPoly(P,N);
      Chrono := HITTimer(TickCount);

    finally
      //----- Restore the default status
      HITNoBreak;
      AbortBtn.Enabled := false;
    end;

    //----- Display the results
    Delete(Count-1);
    Add('Gaussian normal field polynomial');
    Add('--------------------------------');
    Add(GF2HexStr(P,4));
    Add('');

    //----- Display the running time
    Add('Running time = ' + FloatToStrF(Chrono/1000,ffFixed,18,3) + ' seconds');
    Add('');

    //===== Check
    Add('Check (P is normal) -> ' + CHECKMSG[P.IsNormal]);
    Add('');

  finally
    //----- Free the polynomial
    P.Free;
  end;
end;

////////////////////////////////////////////////////////////////////////////////
//
////////////////////////////////////////////////////////////////////////////////

//==============================================================================
// Open the Popup menu
//==============================================================================
procedure TForm1.MenuBtnClick(Sender: TObject);
  var P : TPoint;
begin
  MenuBtn.Enabled := false;
  try
    GetCursorPos(P);
    PopupMenu1.Popup(P.X,P.Y);
  finally
    MenuBtn.Enabled := true;
  end;
end;

//==============================================================================
// Call one of the methods Run
//==============================================================================
procedure TForm1.Run(Sender: TObject; Num: Char);
begin
  if not MenuBtn.Enabled then Exit; //To prevent hot key repetition
  MenuBtn.Enabled := false;
  ClearBtn.Enabled := false;
  CloseBtn.Enabled := false;
  HelpBtn.Enabled := false;
  PopupMenu1.AutoPopup := false;
  try
    Memo1.Lines.Add(#13#10'Example ' + Num);
    Memo1.Lines.Add('---------'#13#10);
    case Num of
      'A': RunA(Sender);
      'B': RunB(Sender);
      'C': RunC(Sender);
      'D': RunD(Sender);
      'E': RunE(Sender);
      'F': RunF(Sender);
      'G': RunG(Sender);
      'H': RunH(Sender);
      'I': RunI(Sender);
    end;
  finally
    MenuBtn.Enabled := true;
    ClearBtn.Enabled := true;
    CloseBtn.Enabled := true;
    HelpBtn.Enabled := true;
    PopupMenu1.AutoPopup := true;
  end;
end;

//==============================================================================
// Call RunA method
//==============================================================================
procedure TForm1.PopItemAClick(Sender: TObject);
begin
  Run(Sender,'A');
end;

//==============================================================================
// Call RunB method
//==============================================================================
procedure TForm1.PopItemBClick(Sender: TObject);
begin
  Run(Sender,'B');
end;

//==============================================================================
// Call RunC method
//==============================================================================
procedure TForm1.PopItemCClick(Sender: TObject);
begin
  Run(Sender,'C');
end;

//==============================================================================
// Call RunD method
//==============================================================================
procedure TForm1.PopItemDClick(Sender: TObject);
begin
  Run(Sender,'D');
end;

//==============================================================================
// Call RunE method
//==============================================================================
procedure TForm1.PopItemEClick(Sender: TObject);
begin
  Run(Sender,'E');
end;

//==============================================================================
// Call RunF method
//==============================================================================
procedure TForm1.PopItemFClick(Sender: TObject);
begin
  Run(Sender,'F');
end;

//==============================================================================
// Call RunG method
//==============================================================================
procedure TForm1.PopItemGClick(Sender: TObject);
begin
  Run(Sender,'G');
end;

//==============================================================================
// Call RunH method
//==============================================================================
procedure TForm1.PopItemHClick(Sender: TObject);
begin
  Run(Sender,'H');
end;

//==============================================================================
// Call RunI method
//==============================================================================
procedure TForm1.PopItemIClick(Sender: TObject);
begin
  Run(Sender,'I');
end;

////////////////////////////////////////////////////////////////////////////////
end.
////////////////////////////////////////////////////////////////////////////////

