////////////////////////////////////////////////////////////////////////////////
// Main unit of the IntDemo project
// Unit of the HIT package for DELPHI 4/5
//------------------------------------------------------------------------------
// Copyright (c) 2000, Marcel Martin. All rights reserved.
////////////////////////////////////////////////////////////////////////////////
unit IntDemoMain;

interface
{$A+,H+}

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, ExtCtrls, Buttons, Menus;

type
  TForm1 = class(TForm)
    Panel1: TPanel;
    Memo1: TMemo;
    AbortBtn: TSpeedButton;
    ClearBtn: TSpeedButton;
    CloseBtn: TSpeedButton;
    HelpBtn: TSpeedButton;
    MenuBtn: TSpeedButton;
    PopupMenu1: TPopupMenu;
    PopItemA: TMenuItem;
    PopItemB: TMenuItem;
    PopItemC: TMenuItem;
    PopItemD: TMenuItem;
    PopItemE: TMenuItem;
    PopItemF: TMenuItem;
    PopItemG: TMenuItem;
    PopItemH: TMenuItem;
    PopItemI: TMenuItem;
    PopItemJ: TMenuItem;
    PopItemK: TMenuItem;
    PopItemL: TMenuItem;
    PopItemM: TMenuItem;
    procedure AbortBtnClick(Sender: TObject);
    procedure ClearBtnClick(Sender: TObject);
    procedure CloseBtnClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure HelpBtnClick(Sender: TObject);
    procedure MenuBtnClick(Sender: TObject);
    procedure PopItemAClick(Sender: TObject);
    procedure PopItemBClick(Sender: TObject);
    procedure PopItemCClick(Sender: TObject);
    procedure PopItemDClick(Sender: TObject);
    procedure PopItemEClick(Sender: TObject);
    procedure PopItemFClick(Sender: TObject);
    procedure PopItemGClick(Sender: TObject);
    procedure PopItemHClick(Sender: TObject);
    procedure PopItemIClick(Sender: TObject);
    procedure PopItemJClick(Sender: TObject);
    procedure PopItemKClick(Sender: TObject);
    procedure PopItemLClick(Sender: TObject);
    procedure PopItemMClick(Sender: TObject);
    procedure Run(Sender: TObject; Num: Char);
    procedure RunA(Sender: TObject);
    procedure RunB(Sender: TObject);
    procedure RunC(Sender: TObject);
    procedure RunD(Sender: TObject);
    procedure RunE(Sender: TObject);
    procedure RunF(Sender: TObject);
    procedure RunG(Sender: TObject);
    procedure RunH(Sender: TObject);
    procedure RunI(Sender: TObject);
    procedure RunJ(Sender: TObject);
    procedure RunK(Sender: TObject);
    procedure RunL(Sender: TObject);
    procedure RunM(Sender: TObject);
  end;

var Form1: TForm1;


implementation
//==============================================================================
{$C-,I-,O+,Q-,R-,W-}
{$R *.DFM}
//==============================================================================
uses
  Integers,
  Integers_FFT,
  Integers_Primes,
  GFP_EllipticCurves,
  GFP_ECBuilder,
  GaussIntegers;

const
  CHECKMSG : array[Boolean] of string[5] = ('ERROR','ok');
  LOGICAL  : array[Boolean] of string[5] = ('FALSE','TRUE');


////////////////////////////////////////////////////////////////////////////////
// TForm1 class
////////////////////////////////////////////////////////////////////////////////

//==============================================================================
// Abort the running of some HIT routines
//==============================================================================
procedure TForm1.AbortBtnClick(Sender: TObject);
begin
  HITAbort;
end;

//==============================================================================
// Clear the editor
//==============================================================================
procedure TForm1.ClearBtnClick(Sender: TObject);
begin
  Memo1.Lines.Clear;
end;

//==============================================================================
// Terminate the application
//==============================================================================
procedure TForm1.CloseBtnClick(Sender: TObject);
begin
  Close;
end;

//==============================================================================
// Create the form
//==============================================================================
procedure TForm1.FormCreate(Sender: TObject);
begin
  Application.HintColor := clYELLOW;

  //----- String format
  //HITStrFormat.SpaceAsPlus := false; //'+' instead of a space
  //HITStrFormat.LowerCased := false; //For hexadecimal strings
end;

//==============================================================================
// Open the HIT.HLP file
//==============================================================================
procedure TForm1.HelpBtnClick(Sender: TObject);
begin
  Application.HelpCommand(HELP_FINDER,0);
end;

////////////////////////////////////////////////////////////////////////////////
// Example A
////////////////////////////////////////////////////////////////////////////////

//==============================================================================
// Generate and display a random integer
//==============================================================================
procedure TForm1.RunA(Sender: TObject);
  var A : TInteger;
begin
  //----- Create A equal to 0
  A := TInteger.Create;
  with Memo1.Lines do
  try
    //----- Random signed integer of at most 166 bits (~ 50 decimal digits)
    IRnd(A,166,false,true);
    //----- Display A
    Add('A (base 10) =' + IStr(A,10,5)); //5-digit blocks
    Add('');
    Add('A (base 16) =' + IStr(A,16,4)); //4-digit blocks
    Add('');
    Add('A (base 8)  =' + IStr(A,8,4));  //4-digit blocks
    Add('');
    Add('A (base 2)  =' + IStr(A,2,8));  //8-digit blocks
    Add('');
  finally
    //----- Free the integer
    A.Free;
  end;
end;

////////////////////////////////////////////////////////////////////////////////
// Example B
////////////////////////////////////////////////////////////////////////////////

//==============================================================================
// Generate and sort a vector of random numbers
//==============================================================================
procedure TForm1.RunB(Sender: TObject);
  var
    A : TInteger;
    V : TIntegerVector;
    i : Integer;
    S : string;
begin
  A := nil;
  V := nil;
  try
    //----- Create A equal to 0
    A := TInteger.Create;
    //----- Create an empty vector
    V := TIntegerVector.Create;

    //===== Compute random coefficients
    for i := 1 to 10 do
    begin
      //----- Random signed integer (random size of at most 200 bits)
      IRnd(A,HITRnd32 mod 201,false,true);
      //----- Append a copy of A to the vector V
      V.AddCell(A);
    end;

    //===== Display the integers to the base 10
    Memo1.Lines.Add('Random vector');
    S := '';
    for i := 0 to V.Size-1 do
      S := S + 'V[' + IntToStr(i) + '] =' + IStr(V[i],10,0) + #13#10;
    Memo1.Lines.Add(S);

    //===== Sort the vector
    V.Sort; //"V.Sort(true)" to sort in absolute values

    //===== Display the sorted list
    Memo1.Lines.Add('Sorted vector');
    S := '';
    for i := 0 to V.Size-1 do
      S := S + 'V[' + IntToStr(i) + '] =' + IStr(V[i],10,0) + #13#10;
    Memo1.Lines.Add(S);

    //===== Get product
    IProduct(A,V,V);

    //===== Display the product
    Memo1.Lines.Add('Sum of V[i] =' + IStr(A,10,0));

  finally
    //----- Free the number
    A.Free;
    //----- Free the vector
    V.Free;
  end;
end;

////////////////////////////////////////////////////////////////////////////////
// Example C
////////////////////////////////////////////////////////////////////////////////

//==============================================================================
// Compute 2^(A-1) modulo A (A is a 664-bit random integer)
//==============================================================================
procedure TForm1.RunC(Sender: TObject);
  var
    S         : string;
    A, E, X   : TInteger;
    BSize, Hw : Longint;
    TickCount : Int64;
    Chrono    : Extended;
begin
  A := nil;
  E := nil;
  X := nil;
  with Memo1.Lines do
  try
    //----- Create integers equal to 0
    A := TInteger.Create;
    E := TInteger.Create;
    X := TInteger.Create;

    //----- Compute a 664-bit random unsigned integer
    BSize := 664; //664 bits ~ 200 decimal digits
    IRnd(A,BSize);
    A.Odd := true; //So that the number is odd

    //----- Display A
    Add('A =' + IStr(A,10,5)); //Base 10, 5-digit blocks

    //----- Display the different sizes of the number
    Add('Size (base 10)   = ' + IntToStr(A.DecimalSize));
    Add('Size (base 2)    = ' + IntToStr(A.BitSize));
    Add('Size (base 2^8)  = ' + IntToStr(A.ByteSize));
    Add('Size (base 2^32) = ' + IntToStr(A.LongSize));
    Add('');

    //----- E = A-1
    ISet(E,A); //Copy A to E
    IDec(E);   //E := E - 1
    //----- X = 2
    ISet(X,2);

    //----- Calculate X := 2^(A-1) mod A
    TickCount := HITTimer;
    IPowMod(X,E,A);
    Chrono := HITTimer(TickCount);

    //----- Display
    Add('2^(A-1) mod A =' + IStr(X,10,5));
    Add('Size (base 10)   = ' + IntToStr(X.DecimalSize));
    Add('Size (base 2)    = ' + IntToStr(X.BitSize));
    Add('Size (base 2^8)  = ' + IntToStr(X.ByteSize));
    Add('Size (base 2^32) = ' + IntToStr(X.LongSize));
    Add('');

    //----- Rare event: Proba(X = 1) ~ 2/A.BitSize
    if X.Is1 then Add('>>> A is pseudoprime to the base 2 <<<'#13#10);

    //----- The work done depends on the size and on the Hamming weight
    //      of the exponent
    Dec(BSize);
    Hw := E.Weight-1;
    Add('The calculations required ' +
         IntToStr(BSize) + ' squares, ' +
         IntToStr(Hw) + ' multiplications and ' +
         IntToStr(BSize+Hw) + ' divisions.');

    //----- Display the running time
    S := FloatToStrF(Chrono,ffFixed,18,3) + ' ms';
    Add('Running time = ' + S);
    Add('');
  finally
    A.Free;
    E.Free;
    X.Free;
  end;
end;

////////////////////////////////////////////////////////////////////////////////
// Example D
////////////////////////////////////////////////////////////////////////////////

//==============================================================================
// Saving and loading integers
//==============================================================================
procedure TForm1.RunD(Sender: TObject);
  const FILENAME = 'random.hit';
  var A, B : TInteger;
begin
  A := nil;
  B := nil;
  with Memo1.Lines do
  try
    //----- Create A equal to 0
    A := TInteger.Create;
    B := TInteger.Create;

    //----- Random signed integer of at most 500 decimal digits
    IRnd(A,1660,false,true);

    //----- Display A
    Add('A =  ' + IStr(A,10,5));
    Add('');

    //----- Save A to the FILENAME file
    A.SaveToFile(FILENAME);

    //----- Load B from the FILENAME file
    B.LoadFromFile(FILENAME);

    //----- Display B
    Add('B =  ' + IStr(B,10,5));
    Add('');

    //----- Check
    Add('Check -> ' + CHECKMSG[IEqu(A,B)]);
    Add('');
  finally
    //----- Free the integers
    A.Free;
    B.Free;
  end;
end;

////////////////////////////////////////////////////////////////////////////////
// Example E
////////////////////////////////////////////////////////////////////////////////

//==============================================================================
// Search the small divisors of a random integer
//==============================================================================
procedure TForm1.RunE(Sender: TObject);
  var
    N, M, F, E : TInteger;
    i, j       : Longint;
    TickCount  : Int64;
    Chrono     : Extended;
begin
  N := nil;
  M := nil;
  F := nil;
  E := nil;

  with Memo1.Lines do
  try
    //----- Create null integers
    N := TInteger.Create;
    M := TInteger.Create;
    F := TInteger.Create;
    E := TInteger.Create;

    //----- Random integer of at most 332 bits (~100 decimal digits)
    IRnd(N,332,false);

    //----- Store a copy of N
    ISet(M,N);

    //----- Display N
    Add('Initial integer:');
    Add('----------------');
    Add('N =' + IStr(N,10,3));
    Add('');

    //----- Search and store all the divisors < 32768 of N
    TickCount := HITTimer;
    ISmallFactors(N,F,E);
    Chrono := HITTimer(TickCount);

    //----- Display the factors, if any
    Add('Small factor(s):');
    Add('----------------');
    if F.LongSize = 0 then Add('None')
    else
    for i := 0 to F.LongSize-1 do
      if E.Long[i] = 1 then Add(IntToStr(F.Long[i]))
      else Add(IntToStr(F.Long[i]) + '^' + IntToStr(E.Long[i]));
    Add('');

    //----- Display the result: N divided by all the found factors
    Add('Residual integer:');
    Add('-----------------');
    Add('R =' + IStr(N,10,3));
    Add('');

    //----- Display the running time
    Add('Running time = ' + FloatToStrF(Chrono,ffFixed,18,3) + ' ms');
    Add('');

    //===== Check
    //----- Rebuild N
    for i := 0 to F.LongSize-1 do
      for j := 1 to E.Long[i] do IMul(N,F.Long[i]);
    //----- Compare to initial value
    Add('Check -> ' + CHECKMSG[IEqu(N,M)]);
    Add('');
  finally
    N.Free;
    M.Free;
    F.Free;
    E.Free;
  end;
end;

////////////////////////////////////////////////////////////////////////////////
// Example F
////////////////////////////////////////////////////////////////////////////////

//==============================================================================
// GCD (Greatest Common Divisor)
//==============================================================================
procedure TForm1.RunF(Sender: TObject);
  var
    A, B, C, D : TInteger;
    S          : string;
    TickCount  : Int64;
    Chrono     : Extended;
begin
  //----- Set to nil so that the "finally" part works in any case
  A := nil;
  B := nil;
  C := nil;
  D := nil;

  with Memo1.Lines do
  try
    //----- Create integers equal to 0
    A := TInteger.Create;
    B := TInteger.Create;

    //----- Random unsigned integers
    IRnd(A,3320);
    IRnd(B,3320);

    //----- Display A
    Add('A =  ' + IStr(A,10,5));
    Add('');
    //----- Display B
    Add('B =  ' + IStr(B,10,5));
    Add('');

    //----- Create C and D equal to A and B
    C := A.Clone;
    D := B.Clone;

    //----- Display the decimal sizes of A and B
    Add('Decimal size of A = ' + IntToStr(A.DecimalSize));
    Add('Decimal size of B = ' + IntToStr(B.DecimalSize));
    Add('');

    Add('Computing GCD ...');

    //----- Compute the GCD of A and B (Euclidean algorithm)
    TickCount := HITTimer;
    while not A.Is0 do
    begin
      //----- (A', B') <- (B mod A, A)
      IMod(B,A);
      ISwp(B,A);
    end;
    Chrono := HITTimer(TickCount);

    //----- Display B (= GCD(A,B))
    Delete(Count-1);
    Add('GCD(A,B) =' + IStr(B,10,0));
    //----- Display the running time
    S := FloatToStrF(Chrono,ffFixed,18,3) + ' ms';
    Add('Running time = ' + S);
    Add('');

    Add('Computing GCD ...');

    //----- Compute the GCD of A and B (with the IGCD routine)
    TickCount := HITTimer;
    IGCD(C,D);
    Chrono := HITTimer(TickCount);

    //----- Display C (= GCD(A,B))
    Delete(Count-1);
    Add('GCD(A,B) =' + IStr(C,10,0));
    //----- Display the running time
    S := FloatToStrF(Chrono,ffFixed,18,3) + ' ms';
    Add('Running time = ' + S);
    Add('');

    // The 2 previous running times show that IGCD is much more efficient than
    // the simple Euclidean algorithm.

  finally
    //----- Free the integers
    A.Free;
    B.Free;
    C.Free;
    D.Free;
  end;
end;

////////////////////////////////////////////////////////////////////////////////
// Example G
////////////////////////////////////////////////////////////////////////////////

//==============================================================================
// Egyptian fractions N/D = 1/d1 + 1/d2 + ... + 1/dk
//------------------------------------------------------------------------------
// This routine doesn't compute fractions exactly like scribes did. The
// ancient Egyptians used not only fractions of type 1/D, they also knew
// the fraction 2/3.
// The implemented algorithm was used by Fibonacci.
//==============================================================================
procedure TForm1.RunG(Sender: TObject);
  const NSIZE = 16; //Binary size of N. Must be small enough. With the given
                    //value, the result can already be pharaonic.
                    //To see that, try with:
                    //  ISet(N,31340);
                    //  ISet(D,2591571893);
  var
    N, D, X : TInteger;
    V       : TIntegerVector; //To check result
    i       : Integer;
begin
  N := nil;
  D := nil;
  X := nil;
  V := nil;
  with Memo1.Lines do
  try
    //----- Create an empty vector
    V := TIntegerVector.Create;

    //----- Create N and D equal to 0
    N := TInteger.Create;
    D := TInteger.Create;

    //----- Generate N and D such that N > 1 and 0 < (N/D) < 1
    //      For perfectionists who wish obtain the real Egyptian
    //      fractions: either generate N and D such that
    //      0 < (N/D) < (2/3), either subtract 2/3 from N/D
    //      when (N/D) >= (2/3).
    repeat
      //----- Random unsigned N of at most NSIZE bits
      IRnd(N,NSIZE,false);
    until ICmp(N,1) > 0; //N must be greater than 1
    repeat
      //----- Random unsigned D of at most 2*NSIZE bits
      IRnd(D,NSIZE+NSIZE,false);
    until ICmp(D,N) > 0; //D must be greater than N

    //----- Display the initial fraction (base 10, 3-digit blocks)
    Add('N =' + IStr(N,10,3));
    Add('D =' + IStr(D,10,3));
    Add('');

    //----- Add N, D and GCD(N,D) to the vector
    V.AddCell(N);
    V.AddCell(D);
    V.AddCell(N);
    IGCD(V[2],D); //V[2] := GCD(N,D).

    //===== Compute and display the "Egyptian" denominators
    Add('N / D =');
    X := TInteger.Create;
    repeat
      //----- Reduce N/D
      ISet(X,N);
      if not IGCD(X,D) then //X := GCD(X,D) and if X <> 1 then
      begin
        IDiv(N,X); //N := N div X
        IDiv(D,X); //D := N div X
      end;

      //---- Exit loop when N = 1
      if N.Is1 then Break;

      //----- Compute the greatest 1/X that is less than N/D
      ISet(X,D);
      IDiv(X,N);
      IInc(X); //X := (N div D) + 1 -> (N/D > 1/X > (N-1)/D)

      //----- Display the new denominator
      Add('1 /' + IStr(X,10,3));
      Add('  +');

      //----- Add the new denominator to the vector V
      V.AddCell(X);

      //----- Compute the residual N'/D' = N/D - 1/X
      IMul(N,X);
      ISub(N,D); //N := N*X-D
      IMul(D,X); //D := D*X
    until false;

    //----- Display the last denominator
    Add('1 /' + IStr(D,10,3));
    Add('');

    //----- Add the last denominator to the vector V
    V.AddCell(D);

    //----- Display the number of denominators
    Add('Denominators: ' + IntToStr(V.Size-3));
    Add('');

    //===== Check
    //----- We start with N/D = 0/1
    N.Set0;
    D.Set1;
    //----- i starts from 3: V[0],V[1] and V[2] contain initial values
    for i := 3 to V.Size-1 do
    begin
      IMul(N,V[i]);
      IAdd(N,D);    //N := N * V[i] + D
      IMul(D,V[i]); //D := D * V[i]
      //----- Reduce the fraction (if possible)
      ISet(X,N);
      if not IGCD(X,D) then //if GCD <> 1 (Here, it cannot be equal to 0)
      begin
        IDiv(N,X); //N := N div X
        IDiv(D,X); //D := N div X
      end;
    end;
    //----- Here, we have the reduced fraction N/D, so if the gcd
    //      of the initial values was different from 1, we have to
    //      multiply N and D by this gcd
    if not V[2].Is1 then
    begin
      IMul(N,V[2]);
      IMul(D,V[2]);
    end;
    Add('Check -> ' + CHECKMSG[IEqu(N,V[0]) and IEqu(D,V[1])]);
    Add('');
  finally
    //----- Free the vector
    V.Free;

    //----- Free the integers
    N.Free;
    D.Free;
    X.Free;
  end;
end;

////////////////////////////////////////////////////////////////////////////////
// Example H
////////////////////////////////////////////////////////////////////////////////

//==============================================================================
// Compare the running times of IMul and TFFT.Mul
// IMul makes use of the Karatsuba's algorithm
// TFFT.Mul makes use of the Fast Fourier Transform
//==============================================================================
procedure TForm1.RunH(Sender: TObject);
  var
    FFT        : TFFT;
    X, Y, Z    : TInteger;
    Len        : Longint;
    S          : string;
    TickCount  : Int64;
    Chrono     : Extended;
begin
  FFT := nil;
  X := nil;
  Y := nil;
  Z := nil;
  with Memo1.Lines do
  try
    //----- Create a TFFT multiplier
    FFT := TFFT.Create;

    //----- Create X and Y equal to 0
    X := TInteger.Create;
    Y := TInteger.Create;

    //----- Get a random integer
    // First, run this routine with the value below, then, increase this
    // value by 1. You will note that the running time of FFT is more
    // than twice as long.
    // That's ok: there are thresholds because of FFT vectors.
    //   8192 * 24 < 393216 <= 16384 * 24, the size of the FFT vectors
    //   is 16384 (to the base 2^24)
    //   16384 * 24 < 393217 <= 32768 * 24, the size of the FFT vectors
    //   is 32768 (to the base 2^24)
    Len := 393216;
    IRnd(X,Len);
    IRnd(Y,Len);

    //----- Create Z equal to X
    Z := X.Clone;

    //----- Allow to abort calculations
    HITBreakWith(Application.ProcessMessages);
    try
      //----- Activate the Abort button (if need be)
      AbortBtn.Enabled := HITBreakOn;

      //----- Calculate X * Y
      TickCount := HITTimer;
      IMul(X,Y);
      Chrono := HITTimer(TickCount);
      //----- Display the running times
      Add('Running times');
      S := FloatToStrF(Chrono,ffFixed,18,3) + ' ms';
      Add('IMul     (Karatsuba)              = ' + S);

      //----- Calculate Z * Y
      TickCount := HITTimer;
      FFT.Mul(Z,Y);
      Chrono := HITTimer(TickCount);
      //----- Display the running times
      S := FloatToStrF(Chrono,ffFixed,18,3) + ' ms';
      Add('TFFT.Mul (Fast Fourier Transform) = ' + S);
      Add('');
    finally
      //----- Restore the default status
      HITNoBreak;
      AbortBtn.Enabled := false;
    end;

    //----- Check (Compare the 2 squares)
    Add('Check -> ' + CHECKMSG[IEqu(X,Z)]);
    Add('');

    //----- Display the memory currently used by the FFT multiplier
    //      ! Do not take in account FFT.InstanceSize
    Add('FFT used memory = ' + IntToStr(FFT.UsedMem div 1024) + ' Kb');
    Add('');

  finally
    //----- Free the FFT multiplier
    FFT.Free;
    //----- Free the integers
    X.Free;
    Y.Free;
    Z.Free;
  end;
end;

////////////////////////////////////////////////////////////////////////////////
// Example I
////////////////////////////////////////////////////////////////////////////////

//==============================================================================
// Modular square roots modulo a prime
//==============================================================================
procedure TForm1.RunI(Sender: TObject);
  const
    PSIZE = 332; //Prime binary size (332 bits ~ 100 decimal digits)
  var
    P, A, R1, R2, S1, S2 : TInteger;
    TickCount            : Int64;
    c1, c2, c3, c4       : Extended;
begin
  P := nil;
  A := nil;
  R1 := nil;
  S1 := nil;
  R2 := nil;
  S2 := nil;

  //----- Allow to abort calculations
  HITBreakWith(Application.ProcessMessages);

  with Memo1.Lines do
  try
    //----- Activate the Abort button (if need be)
    AbortBtn.Enabled := HITBreakOn;
    //----- No signed symbol in displayed numbers
    //      The space (as '+' symbol) is suppressed from strings
    HITStrFormat.Signed := false;

    //----- Create integers equal to 0
    P := TInteger.Create;
    A := TInteger.Create;
    R1 := TInteger.Create;
    R2 := TInteger.Create;
    S1 := TInteger.Create;
    S2 := TInteger.Create;

    Add('Creating prime ...');

    //===== PSIZE-bit prime
    TickCount := HITTimer;
    IRnd(P,PSIZE);       //PSIZE-bit random integer
    IMakePrime(P,[2,3]); //Strong pseudoprime for the bases 2 and 3
    c1 := HITTimer(TickCount);

    Delete(Count-1); //Suppress last line from Memo1.Lines
    Add('Computing quadratic residue ...');

    //===== A such that (A < P) and (Jacobi symbol (A,P) = 1)
    TickCount := HITTimer;
    IRnd(A,PSIZE+PSIZE); //(PSIZE*2)-bit random integer
    IMod(A,P);           //So that A < P
    while IJacobi(A,P) <> 1 do
    begin
      IInc(A);
      if ICmpAbs(A,P) >= 0 then ISub(A,P); //So that A < P (probably never used)
    end;
    c2 := HITTimer(TickCount);

    Delete(Count-1);
    Add('Computing square roots ...');

    //===== R1 and R2 = P - R1
    TickCount := HITTimer;
    ISet(R1,A);
    ISqrRootMod(R1,P);
    ISet(R2,P);
    ISub(R2,R1);
    c3 := HITTimer(TickCount);

    Delete(Count-1);
    Add('Checking ...');

    //===== Squares
    TickCount := HITTimer;
    ISet(S1,R1);
    ISet(S2,R2);
    ISqr(S1);
    IMod(S1,P);
    ISqr(S2);
    IMod(S2,P);
    c4 := HITTimer(TickCount);

    //----- Display
    Delete(Count-1);
    Add('> Prime'#13#10 + IStr(P,10,5));
    Add('');
    Add('> Quadratic residue'#13#10 + IStr(A,10,5));
    Add('');
    Add('> Root 1'#13#10 + IStr(R1,10,5));
    Add('');
    Add('> Root 2'#13#10 + IStr(R2,10,5));
    Add('');
    //----- Check (compare the 2 squares)
    Add('Check 1 -> ' + CHECKMSG[IEqu(S1,A)]);
    Add('Check 2 -> ' + CHECKMSG[IEqu(S2,A)]);
    Add('');

    //----- Running times
    Add('> Running times (milliseconds)');
    Add('Prime ........... ' + FloatToStrF(c1,ffFixed,18,3));
    Add('Quadratic residue ' + FloatToStrF(c2,ffFixed,18,3));
    Add('Square roots .... ' + FloatToStrF(c3,ffFixed,18,3));
    Add('Checks .......... ' + FloatToStrF(c4,ffFixed,18,3));
    Add('');
  finally
    //----- Restore default status
    AbortBtn.Enabled := false;
    HITNoBreak;
    HITStrFormat.Signed := true;
    //----- Free the integers
    P.Free;
    A.Free;
    R1.Free;
    R2.Free;
    S1.Free;
    S2.Free;
  end;
end;

////////////////////////////////////////////////////////////////////////////////
// Example J
////////////////////////////////////////////////////////////////////////////////

//==============================================================================
// RSA cryptosystem.
// Do not use this (very) simplistic demo in an application:
// 1) it is not secure;
// 2) it is not optimized;
//
// 3) RSA is covered by patents in the USA until September 20, 2000
//    -------------------------------------------------------------
//
//==============================================================================
procedure TForm1.RunJ(Sender: TObject);
  //----- Constants to display integers to the base 16, 4-digit blocks
  //      You can use (10,5) or even (8,8) instead of (16,4)
  const
    BASE   = 16;
    BLOCKS = 4;
  var
    P, Q, N, E, D, X, Y, Z : TInteger;
    S                      : string;
    TickCount              : Int64;
    c1, c2, c3, c4         : Extended;
begin
  P := nil;
  Q := nil;
  N := nil;
  E := nil;
  D := nil;
  X := nil;
  Y := nil;
  Z := nil;

  //----- No signed symbol in displayed numbers
  //      The space (as '+' symbol) is suppressed from strings
  HITStrFormat.Signed := false;

  with Memo1.Lines do
  try
    //----- Create integers equal to 0
    P := TInteger.Create;
    Q := TInteger.Create;
    N := TInteger.Create;
    E := TInteger.Create;
    D := TInteger.Create;
    X := TInteger.Create;
    Y := TInteger.Create;
    Z := TInteger.Create;

    Add('Creating modulus ...');

    //===== 512-bit modulus (Product of 2 distinct primes)
    TickCount := HITTimer;
    IRnd(P,254);         //254-bit random integer
    P.Bit[252] := true;  //Set the second leading bit (See below)
    IMakePrime(P,[2,3]); //Strong pseudoprime for the bases 2 and 3
    IRnd(Q,258);
    Q.Bit[256] := true;
    IMakePrime(Q,[2,3]);
    //----- Modulus N = PQ
    ISet(N,P);
    IMul(N,Q);
    c1 := HITTimer(TickCount);

    // Setting the second leading bit of both factors implies that
    // N.BitSize = P.BitSize + Q.BitSize.
    // If the second leading bits were random, the binary size of N could be
    // equal to P.BitSize + Q.BitSize - 1 each other time.

    Delete(Count-1); //Suppress last line from Memo1.Lines
    Add('Creating exponents ...');

    //===== Exponents
    TickCount := HITTimer;
    //----- P := LCM(P-1, Q-1) (Carmichael function)
    //      Most works concerning RSA suggest to take the decryption
    //      exponent D equal to the inverse of E modulo PHI(N) (Euler
    //      PHI-function). It is better to take D equal to the inverse
    //      of E modulo L(N) (Carmichael LAMBDA-function). It generally
    //      minimizes the size of D.
    IDec(P);
    IDec(Q);
    ILCM(P,Q);
    repeat
      //----- Encryption exponent E
      IRnd(E,80);
      E.Odd := true;
      //----- Decryption exponent D := E^(-1) mod LCM(P-1, Q-1)
      ISet(D,E);
      IInvMod(D,P);
    //----- Here D.BitSize > 0 implies GCD(E,L) = 1)
    until IGCD1(D,N);
    c2 := HITTimer(TickCount);

    Delete(Count-1);
    Add('Generating plain text ...');

    //===== Random message X
    //      The "message" is constituted with one repeated digit in order to
    //      make visual checking possible.
    S := StringOfChar(PAnsiChar('123456789abcdef')[HITRnd32 mod (BASE-1)],150);
    ISet(X,S,BASE);
    while ICmpAbs(N,X) <= 0 do IDiv(X,BASE); //X must be less than N

    Delete(Count-1);
    Add('Encrypting ...');

    //===== Encrypt
    TickCount := HITTimer;
    ISet(Y,X);
    IPowMod(Y,E,N);
    c3 := HITTimer(TickCount);

    Delete(Count-1);
    Add('Decrypting ...');

    //===== Decrypt
    TickCount := HITTimer;
    ISet(Z,Y);
    IPowMod(Z,D,N);
    c4 := HITTimer(TickCount);

    //----- Display
    Delete(Count-1);
    Add('> Modulus'#13#10 + IStr(N,BASE,BLOCKS));
    Add('');
    Add('> Encryption exponent'#13#10 + IStr(E,BASE,BLOCKS));
    Add('');
    Add('> Decryption exponent'#13#10 + IStr(D,BASE,BLOCKS));
    Add('');
    Add('> Plain text'#13#10 + IStr(X,BASE,BLOCKS));
    Add('');
    Add('> Encrypted'#13#10 + IStr(Y,BASE,BLOCKS));
    Add('');
    Add('> Decrypted'#13#10 + IStr(Z,BASE,BLOCKS));
    Add('');

    //----- Running times
    Add('> Running times (milliseconds)');
    Add('Modulus ..... ' + FloatToStrF(c1,ffFixed,18,3));
    Add('Exponents ... ' + FloatToStrF(c2,ffFixed,18,3));
    Add('Encryption .. ' + FloatToStrF(c3,ffFixed,18,3));
    Add('Decryption .. ' + FloatToStrF(c4,ffFixed,18,3));
    Add('');
  finally
    //----- Restore default value
    HITStrFormat.Signed := true;
    //----- Free the integers
    P.Free;
    Q.Free;
    N.Free;
    E.Free;
    D.Free;
    X.Free;
    Y.Free;
    Z.Free;
  end;
end;

////////////////////////////////////////////////////////////////////////////////
// Example K
////////////////////////////////////////////////////////////////////////////////

//==============================================================================
// Generation of Elliptic Curves over GF(P) (with "nearly" prime order)
//==============================================================================
procedure TForm1.RunK(Sender: TObject);
  //----- Constants to display numbers
  const
    BASE   = 10;
    BLOCKS = 5;
  //----------------------------------------------------
  // Display the point P
  //----------------------------------------------------
  procedure DisplayPoint(P: TGFPEPoint; ec: TGFPECurve);
  begin
    with Memo1.Lines do
    begin
      if P.IsO then Add('IDENTITY')
      else
      begin
        Add('X =' + IStr(P.X,BASE,BLOCKS));
        Add('Y =' + IStr(P.Y,BASE,BLOCKS));
        if not ec.OnCurve(P) then
          Add('An error occured: The point is not on the curve!?');
      end;
      Add('');
    end;
  end;
  //----------------------------------------------------
  const
    JACOBIAN_COORDINATES = TRUE;
  var
    ec                  : TGFPECurve;
    ep                  : TGFPEPoint;
    M, K, R, A, B, X, Y : TInteger;
    TickCount           : Int64;
    cd                  : array[1..2] of Extended;
begin
  M := nil;
  K := nil;
  R := nil;
  A := nil;
  B := nil;
  X := nil;
  Y := nil;
  ec := nil;
  ep := nil;
  //----- Allow to abort calculations
  HITBreakWith(Application.ProcessMessages);
  with Memo1.Lines do
  try
    //----- Activate the Abort button (if need be)
    AbortBtn.Enabled := HITBreakOn;
    //----- Create integers
    M := TInteger.Create;
    K := TInteger.Create;
    R := TInteger.Create;
    A := TInteger.Create;
    B := TInteger.Create;
    X := TInteger.Create;
    Y := TInteger.Create;
    //----- Create an elliptic curve
    ec := TGFPECurve.Create;
    //----- Create a point
    ep := TGFPEPoint.Create;

    Add('Generating a curve ...');

    //----- Generate a an elliptic curve over GF(P). This curve has
    //      order N = K*R with K smooth and R prime
    TickCount := HITTimer;
    BuildCurve(M,K,R,A,B,X,Y,200,12); //See in GFP_ECBuilder.pas
    cd[1] := HITTimer(TickCount);

    //----- Initialize the curve (modulus + coefs)
    ec.SetModulus(M,false);
    ec.SetCoefs(A,B);

    //----- Set point
    ep.SetXY(X,Y);

    //----- Display
    Delete(Count-1); //Suppress last line from Memo1.Lines
    Add('Curve modulus');
    Add('-------------');
    Add('M =' + IStr(M,BASE,BLOCKS));
    case BASE of
       2: Add('Digits -> ' + IntToStr(M.BitSize));
       8: Add('Digits -> ' + IntToStr((M.BitSize+2) div 3));
      10: Add('Digits -> ' + IntToStr(M.DecimalSize));
      16: Add('Digits -> ' + IntToStr((M.BitSize+3) div 4));
    end;
    Add('');
    Add('Curve coefficients');
    Add('------------------');
    Add('A =' + IStr(A,BASE,BLOCKS));
    Add('B =' + IStr(B,BASE,BLOCKS));
    Add('');
    Add('Curve order (equal to K*R with R prime)');
    Add('---------------------------------------');
    Add('K =' + IStr(K,BASE,BLOCKS));
    Add('R =' + IStr(R,BASE,BLOCKS));
    Add('');
    Add('Point P of order R');
    Add('------------------');
    DisplayPoint(ep,ec);

    Add('Multiplying ...');

    //----- Compute P := P*R
    TickCount := HITTimer;
    ec.Mul(ep,R,JACOBIAN_COORDINATES);
    cd[2] := HITTimer(TickCount);

    //----- Display
    Delete(Count-1); //Suppress last line from Memo1.Lines
    Add('Point Q = P*R (should be equal to IDENTITY)');
    Add('-------------------------------------------');
    DisplayPoint(ep,ec);

    //===== Running times
    Add('Running times (milliseconds):');
    Add('-----------------------------');
    Add('Generation ...... ' + FloatToStrF(cd[1],ffFixed,18,3));
    Add('Multiplication .. ' + FloatToStrF(cd[2],ffFixed,18,3));
    Add('');

  finally
    //----- Restore default status
    AbortBtn.Enabled := false;
    HITNoBreak;
    //----- Free the curve
    ec.Free;
    //----- Free the point
    ep.Free;
    //----- Free the integers
    M.Free;
    K.Free;
    R.Free;
    A.Free;
    B.Free;
    X.Free;
    Y.Free;
  end;
end;

////////////////////////////////////////////////////////////////////////////////
// Example L
////////////////////////////////////////////////////////////////////////////////

//==============================================================================
// Modular inversion over Z[i] (Gaussian integers)
//==============================================================================
procedure TForm1.RunL(Sender: TObject);
  var
    A, B, U   : TGaussInt; //Declared in GaussIntegers.pas
    TickCount : Int64;
    Chrono    : Extended;
    Test      : Boolean;
begin
  A := nil;
  B := nil;
  U := nil;
  with Memo1.Lines do
  try
    //----- Create null Gaussian integers
    A := TGaussInt.Create;
    B := TGaussInt.Create;
    U := TGaussInt.Create;

    //----- Generate 2 random Gaussian integers
    IRnd(A.Re,512,false,true);
    IRnd(A.Im,512,false,true);
    IRnd(B.Re,512,false,true);
    IRnd(B.Im,512,false,true);

    //----- So that Norm(A) < Norm(B)
    ZMod(A,B);
    ZSet(U,A);

    Add('Random A and B (with Norm(A) < Norm(B))');
    Add('---------------------------------------');
    Add('A.r =' + IStr(A.Re,10,5));
    Add('A.i =' + IStr(A.Im,10,5));
    Add('B.r =' + IStr(B.Re,10,5));
    Add('B.i =' + IStr(B.Im,10,5));
    Add('');

    //----- Display "wait" warning
    Add('Computing inverse ...');

    //----- Compute A^(-1) mod B (if possible)
    TickCount := HITTimer;
    Test := ZInvMod(U,B);
    Chrono := HITTimer(TickCount);

    //----- Suppress the "wait" warning
    Delete(Count-1);

    //----- If inversion succeded
    if Test then
    begin
      Add('Inverse of A mod B');
      Add('------------------');
      Add('U.r =' + IStr(U.Re,10,5));
      Add('U.i =' + IStr(U.Im,10,5));
      Add('');

      //===== Check
      ZMul(U,A);
      ZMod(U,B);
      Add('Check (UA mod B = 1) -> ' + CHECKMSG[U.Is1]);
      Add('');

      Add('Inversion running time = ' +
           FloatToStrF(Chrono,ffFixed,18,3) + ' milliseconds');
      Add('');

      // Over Z[i], GCD, factorization, etc. are unique up to a multiplication
      // by a unit (there are 4 units: 1, -1, i and -i).
      // Here we only checked U = 1 because InvMod select between the 4
      // possible inverses the one that makes A*A^(-1) = 1. The 3 other ones
      // can be obtained by multiplying U by -1, i and -i.
    end
    else
    //----- Inversion failed, thus GCD(A,B) <> [1, -1, i, -i]
    begin
      Add('A is not invertible modulo B');
      Add('');

      //----- Display "wait" warning
      Add('Computing GCD ...');

      //----- Compute GCD(A,B)
      ZSet(U,A);
      TickCount := HITTimer;
      ZGCD(U,B);
      Chrono := HITTimer(TickCount);

      //----- Suppress the "wait" warning
      Delete(Count-1);

      //----- Display
      Add('GCD(A,B)');
      Add('--------');
      Add('G.r =' + IStr(U.Re,10,5));
      Add('G.i =' + IStr(U.Im,10,5));
      Add('');

      //===== Check
      ZMod(A,U);
      ZMod(B,U);
      Add('Check (A mod G = 0) -> ' + CHECKMSG[A.Is0]);
      Add('Check (B mod G = 0) -> ' + CHECKMSG[B.Is0]);
      Add('');
      Add('GCD running time = ' +
          FloatToStrF(Chrono,ffFixed,18,3) + ' milliseconds');
      Add('');
    end;
    Add('');

  finally
    //----- Free the Gaussian integers
    A.Free;
    B.Free;
    U.Free;
  end;
end;

////////////////////////////////////////////////////////////////////////////////
// Example M
////////////////////////////////////////////////////////////////////////////////

//==============================================================================
// Create a list containing all the primes less than 2^E
//==============================================================================
procedure TForm1.RunM(Sender: TObject);
  const E = 24; //Should be in 1..30
  var
    L         : TPrimeList;
    i, j      : Longint;
    TickCount : Int64;
    Chrono    : Extended;
begin
  L := nil;
  with Memo1.Lines do
  try
    //----- Create an empty list
    L := TPrimeList.Create;

    //----- Allow to abort calculations
    HITBreakWith(Application.ProcessMessages);
    try
      //----- Activate the Abort button (if need be)
      AbortBtn.Enabled := HITBreakOn;

      Add('Generating the prime list ...');

      //----- Generate the list
      //      Pi(x) = number of primes less than or equal to x
      TickCount := HITTimer;
      L.Count := Pi2Pow[E];
      Chrono := HITTimer(TickCount); //Chrono in milliseconds

    finally
      //----- Restore the default status
      HITNoBreak;
      AbortBtn.Enabled := false;
    end;

    //----- Display infos
    Delete(Count-1);
    Add('The list contains ' + IntToStr(L.Count) + ' primes (Mem size = ' +
        IntToStr(L.Count * 4) +  ' bytes)');
    Add('The greatest prime of the list is ' + IntToStr(L.LastPrime));
    Add('');

    //----- Display the running time
    Add('Creation time = ' + FloatToStrF(Chrono,ffFixed,18,3) + ' milliseconds');
    Add('On average, HIT found ' +
         FloatToStrF(L.Count / Chrono,ffFixed,18,3) + ' primes by millisecond');
    Add('');

    //----- Display some primes
    j := (Longint(HITRnd32 and $7fffffff) mod L.Count) + 1;
    for i := 0 to 9 do
    begin
      Add('Prime[' + IntToStr(j) + '] = ' + IntToStr(L[j]) +
          '   Check (Is prime) -> ' + LOGICAL[ITruePrime(L[j])]);
      Inc(j);
      if j > L.Count then j := 1;
    end;
    Add('');
  finally
    //----- Free the list
    L.Free;
  end;
end;

////////////////////////////////////////////////////////////////////////////////
//
////////////////////////////////////////////////////////////////////////////////

//==============================================================================
// Open the Popup menu
//==============================================================================
procedure TForm1.MenuBtnClick(Sender: TObject);
  var P : TPoint;
begin
  MenuBtn.Enabled := false;
  try
    GetCursorPos(P);
    PopupMenu1.Popup(P.X,P.Y);
  finally
    MenuBtn.Enabled := true;
  end;
end;

//==============================================================================
// Call one of the methods Run
//==============================================================================
procedure TForm1.Run(Sender: TObject; Num: Char);
begin
  if not MenuBtn.Enabled then Exit; //To prevent hot key repetition
  MenuBtn.Enabled := false;
  ClearBtn.Enabled := false;
  CloseBtn.Enabled := false;
  HelpBtn.Enabled := false;
  PopupMenu1.AutoPopup := false;
  try
    Memo1.Lines.Add(#13#10'Example ' + Num);
    Memo1.Lines.Add('---------'#13#10);
    case Num of
      'A': RunA(Sender);
      'B': RunB(Sender);
      'C': RunC(Sender);
      'D': RunD(Sender);
      'E': RunE(Sender);
      'F': RunF(Sender);
      'G': RunG(Sender);
      'H': RunH(Sender);
      'I': RunI(Sender);
      'J': RunJ(Sender);
      'K': RunK(Sender);
      'L': RunL(Sender);
      'M': RunM(Sender);
    end;
  finally
    MenuBtn.Enabled := true;
    ClearBtn.Enabled := true;
    CloseBtn.Enabled := true;
    HelpBtn.Enabled := true;
    PopupMenu1.AutoPopup := true;
  end;
end;

//==============================================================================
// Call RunA method
//==============================================================================
procedure TForm1.PopItemAClick(Sender: TObject);
begin
  Run(Sender,'A');
end;

//==============================================================================
// Call RunB method
//==============================================================================
procedure TForm1.PopItemBClick(Sender: TObject);
begin
  Run(Sender,'B');
end;

//==============================================================================
// Call RunC method
//==============================================================================
procedure TForm1.PopItemCClick(Sender: TObject);
begin
  Run(Sender,'C');
end;

//==============================================================================
// Call RunD method
//==============================================================================
procedure TForm1.PopItemDClick(Sender: TObject);
begin
  Run(Sender,'D');
end;

//==============================================================================
// Call RunE method
//==============================================================================
procedure TForm1.PopItemEClick(Sender: TObject);
begin
  Run(Sender,'E');
end;

//==============================================================================
// Call RunF method
//==============================================================================
procedure TForm1.PopItemFClick(Sender: TObject);
begin
  Run(Sender,'F');
end;

//==============================================================================
// Call RunG method
//==============================================================================
procedure TForm1.PopItemGClick(Sender: TObject);
begin
  Run(Sender,'G');
end;

//==============================================================================
// Call RunH method
//==============================================================================
procedure TForm1.PopItemHClick(Sender: TObject);
begin
  Run(Sender,'H');
end;

//==============================================================================
// Call RunI method
//==============================================================================
procedure TForm1.PopItemIClick(Sender: TObject);
begin
  Run(Sender,'I');
end;

//==============================================================================
// Call RunJ method
//==============================================================================
procedure TForm1.PopItemJClick(Sender: TObject);
begin
  Run(Sender,'J');
end;

//==============================================================================
// Call RunK method
//==============================================================================
procedure TForm1.PopItemKClick(Sender: TObject);
begin
  Run(Sender,'K');
end;

//==============================================================================
// Call RunL method
//==============================================================================
procedure TForm1.PopItemLClick(Sender: TObject);
begin
  Run(Sender,'L');
end;

//==============================================================================
// Call RunM method
//==============================================================================
procedure TForm1.PopItemMClick(Sender: TObject);
begin
  Run(Sender,'M');
end;

////////////////////////////////////////////////////////////////////////////////
end.
////////////////////////////////////////////////////////////////////////////////

